#include <string>
#include <vector>
#include <map>

#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <unistd.h>
#include <stdarg.h>
#include <vector>

#include <assert.h>

#include <gd.h>
#include <libxslt/transform.h>
#include <libxml/xpath.h>

#include "nesvideos-piece.hh"
#include "quantize.hh"
#include "lanczos.hh"
#include "rgbtorgb.hh"
#include "fparser.hh"
#include <iostream>

namespace
{
    /* CONFIGURATION IS IN hudmaker.xml -- NO NEED TO MODIFY THESE DEFAULTS */

    /* Input stream specs */
    std::string VideoInputFileName="/dev/null";
    unsigned    IN_BPP=16, OUT_BPP=16;
    unsigned    ORIGWIDTH=320, ORIGHEIGHT=200;
    unsigned    WIDTH=320, HEIGHT=200;
    // FPS, represented as a ratio of two integers:
    unsigned long long FPSHI=1,    FPSLO=1;
    unsigned long long FPSNEWHI=1, FPSNEWLO=1;
    // Which resolution subtitles will be rendered at
    // (they will be downscaled to WIDTH&HEIGHT):
    unsigned    SUBTITLE_WIDTH=320, SUBTITLE_HEIGHT=200;
    std::string SUBTITLE_FONT="Helvetica";
    /**** AUDIO ****/
    unsigned RATE=11025, CHANNELS=1, BITS=16;
    std::string AudioInputFileName="/dev/null";
    /**** LOGO ****/
    unsigned DELAY_BEGIN=0, FIRST_FRAME=0, LAST_FRAME=0x7FFFFFFFu;
    // Input will be delayed by this amount
    // (to allow for logo to display without blocking input):
    std::string EncodingRecipe="cat >/dev/null";

    struct VideoOverlay
    {
        std::string filename_pattern;
        long    starting_frame;
        long    num_frames;
        long    fade_length;
        long    fade_width;
    };
}

class Overlays
{
    std::vector<VideoOverlay> overlays;
public:
    void AddOverlay(const VideoOverlay& ov)
    {
        overlays.push_back(ov);
    }
    typedef std::vector< std::pair<std::string, double> > specs_t;
    static std::string MakeTextFile(const specs_t& specs)
    {
        std::string found_fn;
        static unsigned textcounter = 0;
        char textfn[64];
        #pragma omp critical(textfile_cache)
      {
        std::sprintf(textfn, "text_%d_%u.png", getpid(), textcounter++);

        if(true) /* Cache already-generated files */
        {
            typedef std::map<specs_t, std::string> produced_files_t;
            static produced_files_t produced_files;

            produced_files_t::iterator produced_files_iterator =
                produced_files.lower_bound(specs);
            if(produced_files_iterator != produced_files.end()
            && produced_files_iterator->first == specs)
            {
                std::fprintf(stderr, "Reusing file %s\n",
                    produced_files_iterator->second.c_str());
                found_fn = produced_files_iterator->second;
            }
            else
                produced_files.insert(produced_files_iterator,
                    std::make_pair(specs, textfn) );
        }
      }
        if(!found_fn.empty()) return found_fn;

        gdImagePtr im = gdImageCreateTrueColor(SUBTITLE_WIDTH, SUBTITLE_HEIGHT);
        gdImagePtr shadow = gdImageCreateTrueColor(SUBTITLE_WIDTH, SUBTITLE_HEIGHT);

        const unsigned TransparentBackground = 0x7F808080;

        gdImageAlphaBlending(im, 0);
        gdImageAlphaBlending(shadow, 0);
        gdImageFill(im,     0,0, TransparentBackground);
        gdImageFill(shadow, 0,0, TransparentBackground);
        gdImageAlphaBlending(im, 1);
        gdImageAlphaBlending(shadow, 1);

        int x = 6   * SUBTITLE_WIDTH  / 256;
        int y = 190 * SUBTITLE_HEIGHT / 240;

        const char* font = SUBTITLE_FONT.c_str();

        for(size_t a=0; a<specs.size(); ++a)
        {
            double size     = specs[a].second;
            const char* str = specs[a].first.c_str();

            int dropshadow_width  = SUBTITLE_WIDTH/200;
            int dropshadow_height = SUBTITLE_HEIGHT/200;
            if(size < 3.0)
            {
                dropshadow_width  /= 3;
                dropshadow_height /= 3;
            }
            if(dropshadow_width < 2) dropshadow_width = 2;
            if(dropshadow_height < 2) dropshadow_height = 2;

            size *= (SUBTITLE_HEIGHT/240.0);

            std::fprintf(stderr, "Rendering '%s' at size %g to %d,%d with %s\n", str, size, x,y, font);

            int textheight, yskew;
            GetTextBoundingRect(font, size, str,
                textheight, yskew);

            int brect[8];

            for(int xo=0; xo<dropshadow_width; ++xo)
            for(int yo=0; yo<dropshadow_height; ++yo)
                gdImageStringFT(shadow, brect, 0x00010101,
                    const_cast<char*> (font),
                    size, 0.0, x+xo,y+yo+yskew, const_cast<char*> ( str )
                               );
            gdImageStringFT(im, brect, 0xFFFFFF,
                const_cast<char*> (font),
                size, 0.0, x,y+yskew, const_cast<char*> ( str )
                           );
            /*gdImageRectangle(im,
                x,y,
                SUBTITLE_WIDTH-20,y+textheight,
                0xFF0000);*/
            y += textheight + 2 * (SUBTITLE_HEIGHT/240);
        }

        /* Overlay the rendered text over top of the shadows */
        gdImageCopy(shadow, im, 0,0, 0,0, SUBTITLE_WIDTH,SUBTITLE_HEIGHT);
        gdImageDestroy(im);

        std::vector<Triplet<unsigned char> > input_frame (SUBTITLE_WIDTH*SUBTITLE_HEIGHT);
        std::vector<Triplet<LanczosFloat>  > scaled_frame;

        for(unsigned p=0,y=0; y<SUBTITLE_HEIGHT; ++y)
            for(unsigned x=0; x<SUBTITLE_WIDTH; ++x,++p)
            {
                unsigned c = gdImageGetTrueColorPixel(shadow, x,y);
                input_frame[p] = Triplet<unsigned char> (c, c>>8, c>>16, c>>24);
            }

        if(SUBTITLE_WIDTH == WIDTH && SUBTITLE_HEIGHT == HEIGHT)
        {
            FILE* out = std::fopen(textfn, "wb");
            gdImageSaveAlpha(shadow, 1);
            gdImagePng(shadow, out);
            std::fclose(out);
            gdImageDestroy(shadow);
        }
        else
        {
            gdImageDestroy(shadow);
            LanczosScale(SUBTITLE_WIDTH,SUBTITLE_HEIGHT,
                         WIDTH,HEIGHT,
                         input_frame,scaled_frame);
            im = gdImageCreateTrueColor(WIDTH, HEIGHT);
            gdImageAlphaBlending(im, 0);
            gdImageSaveAlpha(im, 1);
            for(unsigned p=0,y=0; y<HEIGHT; ++y)
                for(unsigned x=0; x<WIDTH; ++x,++p)
                {
                    const int o16 = (x+4*y)%16;
                    int r = Quantize4x4<0xFF>     (o16, scaled_frame[p][0]); if(r<0)r=0; if(r>255)r=255;
                    int g = Quantize4x4<0xFF>     (o16, scaled_frame[p][1]); if(g<0)r=0; if(g>255)r=255;
                    int b = Quantize4x4<0xFF>     (o16, scaled_frame[p][2]); if(b<0)r=0; if(b>255)r=255;
                    int a = Quantize4x4<0x7F,0x7F>(o16, scaled_frame[p][3]); if(a<0)r=0; if(a>127)r=127;
                    unsigned c = (a<<24) | (r<<16) | (g<<8) | b;
                    gdImageSetPixel(im, x,y, c);
                }
            FILE* out = std::fopen(textfn, "wb");
            gdImagePng(im, out);
            std::fclose(out);
            gdImageDestroy(im);
        }

        std::fprintf(stderr, "Produced file %s\n", textfn);

        return textfn;
    }

    void AddTo(char* VidBuf, long framecounter)
    {
        std::vector<unsigned char> logodata;

        for(size_t a=0; a<overlays.size(); ++a)
        {
            const VideoOverlay& ov = overlays[a];
            long first_affected_frame = overlays[a].starting_frame
                                      - overlays[a].fade_length;
            long end_affected_frame = overlays[a].starting_frame
                                    + overlays[a].num_frames
                                    + overlays[a].fade_length;
            if(framecounter < first_affected_frame
            || framecounter >= end_affected_frame) continue;

            /* Needs an overlay. Make sure we've got a 24bpp image to work with. */
            if(logodata.empty())
                ConvertInputVideoBPP(VidBuf, logodata);

            if(Prefer32bitTempFormat())
                DoOverlay<4>(ov, logodata, framecounter);
            else
                DoOverlay<3>(ov, logodata, framecounter);
        }

        /* If we're doing a BPP change, we will need to convert the input image in any case */
        if(logodata.empty() && IN_BPP != OUT_BPP)
            ConvertInputVideoBPP(VidBuf, logodata);

        /* If we have image data in logodata[],
         * store it to VidBuf with appropriate BPP */
        if(!logodata.empty())
            ConvertOutputVideoBPP(VidBuf, logodata);
    }
private:
    static void Overlay24With24
        (unsigned char* target, const unsigned char* source, int howmuch)
    {
        for(unsigned c=0; c<3; ++c)
            target[c] += ((int)(source[c] - target[c])) * howmuch / 255;
    }

    static bool Prefer32bitTempFormat()
    {
        return (OUT_BPP == 32 || OUT_BPP == 12 || OUT_BPP == 17
                            /*|| OUT_BPP == 15 || OUT_BPP == 16 // - no advantage*/
               )
            && (IN_BPP == 32 || IN_BPP == 12 || IN_BPP == 17
                             || IN_BPP == 15 || IN_BPP == 16
               );
    }

    static void ConvertInputVideoBPP
        (const char* VidBuf, std::vector<unsigned char>& logodata)
    {
        /* RGB24 or RGB32 filled with black. */
        logodata.resize(WIDTH*HEIGHT * (Prefer32bitTempFormat() ? 4 : 3));
        if(IN_BPP == 16)
        {
            if(Prefer32bitTempFormat())
                Convert16To32Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, false);
            else
                Convert16To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, false);
        }
        else if(IN_BPP == 15)
        {
            if(Prefer32bitTempFormat())
                Convert15To32Frame(VidBuf, &logodata[0], WIDTH*HEIGHT);
            else
                Convert15To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT);
        }
        else if(IN_BPP == 17)
        {
            if(Prefer32bitTempFormat())
                Convert_YUY2To32Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, WIDTH);
            else
                Convert_YUY2To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, WIDTH);
        }
        else if(IN_BPP == 12)
        {
            if(Prefer32bitTempFormat())
                Convert_I420To32Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, WIDTH);
            else
                Convert_I420To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, WIDTH);
        }
        else if(IN_BPP == 32)
        {
            if(Prefer32bitTempFormat())
                std::memcpy(&logodata[0], VidBuf, WIDTH*HEIGHT*4);
            else
                Convert32To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT);
        }
        else if(IN_BPP == 24)
            std::memcpy(&logodata[0], VidBuf, WIDTH*HEIGHT*3);
        else
        {
            std::fprintf(stderr, "hudmaker: I do not know how to handle input bpp %u\n", IN_BPP);
            std::abort();
        }
    }

    static void ConvertOutputVideoBPP
        (char* VidBuf, const std::vector<unsigned char>& logodata)
    {
        if(OUT_BPP == 16)
        {
            if(Prefer32bitTempFormat())
                Convert32To16Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
            else
                Convert24To16Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        }
        else if(OUT_BPP == 15)
        {
            if(Prefer32bitTempFormat())
                Convert32To15Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
            else
                Convert24To15Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        }
        else if(OUT_BPP == 17)
        {
            if(Prefer32bitTempFormat())
                Convert32To_YUY2Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
            else
                Convert24To_YUY2Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        }
        else if(OUT_BPP == 12)
        {
            if(Prefer32bitTempFormat())
                Convert32To_I420Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
            else
                Convert24To_I420Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        }
        else if(OUT_BPP == 8)
            Convert24To8Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        else if(OUT_BPP == 24)
            std::memcpy(VidBuf, &logodata[0], logodata.size());
        else if(OUT_BPP == 32 && Prefer32bitTempFormat())
            std::memcpy(VidBuf, &logodata[0], logodata.size());
        else
        {
            std::fprintf(stderr, "hudmaker: I do not know how to handle output bpp %u\n", OUT_BPP);
            std::abort();
        }
    }

    template<unsigned OutPixStride>
    void DoOverlay(
        const VideoOverlay& overlay,
        std::vector<unsigned char>& logodata,
        long framecounter)
    {
        /* Determine where we are in the animation sequence */
        long vis_begin = overlay.starting_frame;
        long vis_end   = vis_begin + overlay.num_frames;
        long fadein_begin  = vis_begin - overlay.fade_length;
      //long fadeout_end   = vis_end   + overlay.fade_length;
        long relative_frame = framecounter < vis_begin ? 0 : framecounter;

        /* Find the image file to read and get its size */
        char fnbuf[4096];
        std::sprintf( fnbuf, overlay.filename_pattern.c_str(),
            WIDTH, HEIGHT, (int) relative_frame );
        static std::string oldfn;
        static gdImagePtr im;

        if(oldfn != fnbuf)
        {
            if(im) gdImageDestroy(im);
            oldfn = fnbuf;
            FILE* fp = std::fopen(fnbuf, "rb");
            if(!fp) { std::perror(fnbuf); return; }
            static char png_read_buffer[1048576];
            std::setvbuf(fp, png_read_buffer, _IOFBF, sizeof(png_read_buffer));
            im = gdImageCreateFromPng(fp);
            std::fclose(fp);
        }
        if(!im) {
            std::fprintf(stderr, "'%s': Failed to read as png file\n", fnbuf);
            gdImageDestroy(im); return; }
        if(!gdImageTrueColor(im)) {
            std::fprintf(stderr, "'%s': Only true color images are supported\n", fnbuf);
            gdImageDestroy(im); return; }
        unsigned new_width = gdImageSX(im);
        unsigned new_height= gdImageSY(im);
        if(new_width != WIDTH || new_height != HEIGHT)
        {
            if(new_height < HEIGHT || new_height > HEIGHT+20)
            std::fprintf(stderr, "'%s': ERROR, expected %ux%u, got %ux%u\n", fnbuf,
                WIDTH, HEIGHT,
                new_width, new_height);
        }

        /* Determine the opacity factors for each column */
        std::vector<int> howmuch_x( WIDTH, 255 );

        if(framecounter < vis_begin)
        {
            double fade_position = (framecounter - fadein_begin) / double(overlay.fade_length);
            // 0 <= fade_position < 1
            long fade_x_end   = (WIDTH + overlay.fade_width) * fade_position;
            long fade_x_begin = fade_x_end - overlay.fade_width;
            for(long x=0; x<WIDTH; ++x)
                if(x < fade_x_begin)
                    howmuch_x[x] = 255;
                else if(x >= fade_x_end)
                    howmuch_x[x] = 0;
                else
                    howmuch_x[x] = 255 - (x-fade_x_begin) * 255 / overlay.fade_width;
            // fade-edge goes like this:
            //   255 -> 255..0 -> 0
        }
        else if(framecounter >= vis_end)
        {
            double fade_position = (framecounter - vis_end) / double(overlay.fade_length);
            // 0 <= fade_position < 1
            long fade_x_end   = (WIDTH + overlay.fade_width) * fade_position;
            long fade_x_begin = fade_x_end - overlay.fade_width;
            for(long x=0; x<WIDTH; ++x)
                if(x < fade_x_begin)
                    howmuch_x[x] = 0;
                else if(x >= fade_x_end)
                    howmuch_x[x] = 255;
                else
                    howmuch_x[x] = (x-fade_x_begin) * 255 / overlay.fade_width;
            // fade-edge goes like this:
            //   0 -> 0..255 -> 255
        }

        /* Superimpose the image over the background. */
        #pragma omp parallel for schedule(static) num_threads(3)
        for(unsigned y=0; y<HEIGHT; ++y)
        {
            unsigned char pixbuf[4] = {0,0,0,0};
            for(unsigned x = 0; x < WIDTH; ++x)
            {
                int color = gdImageTrueColorPixel(im, x,y);
                int howmuch = ((127-gdTrueColorGetAlpha(color)) * howmuch_x[x]) / 127;
                if(howmuch == 0) continue;

                pixbuf[2] = gdTrueColorGetRed(color);
                pixbuf[1] = gdTrueColorGetGreen(color);
                pixbuf[0] = gdTrueColorGetBlue(color);
                Overlay24With24(&logodata[(y*WIDTH+x)*OutPixStride], pixbuf, howmuch);
            }
        }
        //gdImageDestroy(im);
    }


    // Get the bounding rectangle of the given text.
    // Output slots of importance:
    //    brect[0] = left
    //    brect[2] = right
    //    brect[5] = upper
    //    brect[1] = below
    //    brect[7] = value to add to gdImageStringFT()'s y parameter in
    //               order to render the text so that no pixel goes above y
    static void GetTextBoundingRect(
        const char* font,
        const double size,
        const char* str,
        int& textheight,
        int& yskew)
    {
        int brect[8]={-1,-1,-1,-1,-1,-1,-1,-1};
        char*tmp = gdImageStringFT(NULL, brect, 0,
            const_cast<char*> (font),
            size, 0.0, 0, 200, const_cast<char*> ( str )
                       );
        std::fprintf(stderr, "  brect(%s): %d,%d, %d,%d, %d,%d, %d,%d: %s\n",
            font,
            brect[0],brect[1],
            brect[2],brect[3],
            brect[4],brect[5],
            brect[6],brect[7],
            tmp ? tmp : "ok");

        int text_x1 = brect[0];
        int text_x2 = brect[2];
        int text_y1 = brect[5]; // above 0          (i.e. -59)
        int text_y2 = brect[1]; // slightly below 0 (i.e. +3)

        gdImagePtr im = gdImageCreateTrueColor(text_x2, text_y2);
        gdImageFill(im, 0,0, 0x000000);

        gdImageStringFT(im, brect, 0xFFFFFF,
            const_cast<char*> (font),
            size, 0.0, 0, 200, const_cast<char*> ( str )
                       );
        // Find the _true_ boundaries of the text
        int res_x1 = text_x2, res_x2 = text_x1;
        int res_y1 = text_y2, res_y2 = text_y1;
        for(int x=text_x1; x<=text_x2; ++x)
            for(int y=text_y1; y<=text_y2; ++y)
                if(gdImageGetPixel(im, x,y) != 0)
                {
                    if(x < res_x1) res_x1 = x;
                    if(y < res_y1) res_y1 = y;
                    if(x > res_x2) res_x2 = x;
                    if(y > res_y2) res_y2 = y;
                }
         gdImageDestroy(im);

         textheight = res_y2 - res_y1;
         yskew      = 200-res_y1;

         std::fprintf(stderr, "  Bounding box determined: %d,%d .. %d,%d; height=%d, yskew=%d\n",
            res_x1,res_y1, res_x2,res_y2,
            textheight,yskew);
    }
} ov;

class Config
{
public:
    void Parse()
    {
        xmlSubstituteEntitiesDefault(1);
        xmlDocPtr         doc = xmlParseFile("hudmaker.xml");
        xsltStylesheetPtr cur = xsltParseStylesheetFile(BAD_CAST "hudmaker.xsl");
        xmlDocPtr res = xsltApplyStylesheet(cur, doc, 0);
        xsltFreeStylesheet(cur);
        ParseConfig(res);
        xmlFreeDoc(res);
        xmlFreeDoc(doc);
        xsltCleanupGlobals();
        xmlCleanupParser();
    }
private:
    class xpwrap
    {
    private:
        xmlXPathObjectPtr obj;
    private:
        xpwrap(const xpwrap&);
        void operator=(const xpwrap&);
        xpwrap& operator=(int); xpwrap& operator=(const void*);
    public:
        xpwrap() : obj(0) { }
        ~xpwrap() {  if(obj) xmlXPathFreeObject(obj); }
        explicit xpwrap(xmlXPathObjectPtr o) : obj(o) { }

        xpwrap& operator=(xmlXPathObjectPtr o)
        {
            if(obj)  xmlXPathFreeObject(obj);
            obj = o;
            return *this;
        }
        operator bool() const { return obj; }
        xmlXPathObjectPtr operator-> () const { return obj; }

        std::string Get(const char*)
        {
            if(obj)
            {
                obj = xmlXPathConvertString(obj);
                if(obj)
                {
                    const char* input = (const char*) obj->stringval;
                    std::string result;
                    enum { wsIgnore, wsNone } wsmode = wsIgnore;
                    for(; *input; ++input)
                    {
                        if(*input==' '||*input=='\t'||*input=='\r'
                        || *input=='\v'||*input=='\n')
                        {
                            if(wsmode != wsIgnore)
                            {
                                result += ' ';
                                wsmode = wsIgnore;
                            }
                        }
                        else
                        {
                            result += *input;
                            wsmode = wsNone;
                        }
                    }
                    return result;
                }
            }
            return "?";
        }
        double Get(double)
        {
            std::string formula = Get( (const char*)0 );
            FunctionParser parser;
            parser.Parse(formula.c_str(), "");
            parser.AddConstant("WIDTH",  WIDTH);
            parser.AddConstant("HEIGHT", HEIGHT);
            parser.AddConstant("INFPS",    double(FPSHI)    / double(FPSLO));
            parser.AddConstant("OUTFPS",   double(FPSNEWHI) / double(FPSNEWLO));
            return parser.Eval(0);
        }
    };
    void ParseConfig(xmlDocPtr cfg)
    {
        xmlXPathContextPtr xpathCtx = xmlXPathNewContext(cfg);
        ParseConfig(xpathCtx);
        xmlXPathFreeContext(xpathCtx);
    }
    void ParseConfig(xmlXPathContextPtr ctx)
    {
        Get<const char*> (VideoInputFileName, "/hudmaker/input[@type='video']/file", ctx);
        Get<double>      (IN_BPP,             "/hudmaker/input[@type='video']/bpp",  ctx);
        Get<double>      (ORIGWIDTH,          "/hudmaker/input[@type='video']/width", ctx);
        Get<double>      (ORIGHEIGHT,         "/hudmaker/input[@type='video']/height", ctx);
        Get<double>      (FPSHI,              "/hudmaker/input[@type='video']/fps/nom", ctx);
        Get<double>      (FPSLO,              "/hudmaker/input[@type='video']/fps/den", ctx);
        Get<double>      (OUT_BPP,            "/hudmaker/output/bpp", ctx);
        Get<double>      (WIDTH,              "/hudmaker/output/width", ctx);
        Get<double>      (HEIGHT,             "/hudmaker/output/height", ctx);
        Get<double>      (FPSNEWHI,           "/hudmaker/output/fps/nom", ctx);
        Get<double>      (FPSNEWLO,           "/hudmaker/output/fps/den", ctx);
        Get<double>      (DELAY_BEGIN,        "/hudmaker/output/delay", ctx);
        Get<double>      (FIRST_FRAME,        "/hudmaker/output/firstframe", ctx);
        Get<double>      (LAST_FRAME,         "/hudmaker/output/lastframe", ctx);
        Get<const char*> (EncodingRecipe,     "/hudmaker/output/recipe", ctx);
        Get<const char*> (AudioInputFileName, "/hudmaker/input[@type='audio']/file", ctx);
        Get<double>      (RATE,               "/hudmaker/input[@type='audio']/rate",  ctx);
        Get<double>      (CHANNELS,           "/hudmaker/input[@type='audio']/channels", ctx);
        Get<double>      (BITS,               "/hudmaker/input[@type='audio']/bits", ctx);
        Get<double>      (SUBTITLE_WIDTH,     "/hudmaker/overlays/width", ctx);
        Get<double>      (SUBTITLE_HEIGHT,    "/hudmaker/overlays/height", ctx);
        Get<const char*> (SUBTITLE_FONT,      "/hudmaker/overlays/font", ctx);

        xpwrap obj ( xmlXPathEvalExpression(BAD_CAST "/hudmaker/overlays/overlay", ctx) );
        int max = obj ? (obj->nodesetval ? obj->nodesetval->nodeNr : 0) : 0;
      #pragma omp parallel for ordered schedule(dynamic,1) num_threads(16)
        for(int i=0; i<max; ++i)
        {
            xmlNodePtr n = obj->nodesetval->nodeTab[i];
            xmlXPathContextPtr ctx1 = xmlXPathNewContext(ctx->doc);
            ctx1->node = n;
            VideoOverlay o;
            Get<const char*> (o.filename_pattern, "file", ctx1);
            Get<double> (o.starting_frame, "startframe", ctx1);
            Get<double> (o.num_frames,  "numframes",  ctx1);
            Get<double> (o.fade_length, "fadelength", ctx1);
            Get<double> (o.fade_width,  "fadewidth",  ctx1);
            xpwrap obj2 ( xmlXPathEvalExpression(BAD_CAST "text", ctx1) );
            if(obj2)
            {
                Overlays::specs_t specs;
                int i2=0;
                while(xmlNodePtr n2 = xmlXPathNodeSetItem(obj2->nodesetval, i2))
                {
                    xmlXPathContextPtr ctx2 = xmlXPathNewContext(ctx->doc);
                    ctx2->node = n2;
                    double      size;
                    std::string s;
                    Get<double>     (size, "@size", ctx2);
                    Get<const char*>(s,    ".",     ctx2);
                    specs.push_back( std::make_pair(s, size) );
                    xmlXPathFreeContext(ctx2);
                    ++i2;
                }
                if(!specs.empty())
                    o.filename_pattern = Overlays::MakeTextFile(specs);
            }
          #pragma omp ordered
          {
            ov.AddOverlay(o);
          }
            xmlXPathFreeContext(ctx1);
        }
    }
    template<typename T2, typename T>
    void Get(T& target, const char* path, xmlXPathContextPtr ctx)
    {
        xpwrap obj ( xmlXPathEvalExpression(BAD_CAST path, ctx) );
        if(obj)
            target = obj.Get( T2() );
        else
            std::cout << "Config: Could not find: " << path << "\n";
        std::cout << std::flush;
    }
} Config;



int main(int argc, char**argv)
{
    gdFTUseFontConfig(1);
    Config.Parse();

    LoggingEnabled = 2;
    NESVideoSetVideoCmd(EncodingRecipe.c_str());

    FILE* ai = std::fopen(AudioInputFileName.c_str(), "rb");
    FILE* vi = std::fopen(VideoInputFileName.c_str(), "rb");
    if(!ai && RATE) { std::perror(AudioInputFileName.c_str()); }
    if(!vi) { std::perror(VideoInputFileName.c_str()); }
    if(!ai || !vi) return -1;

    const unsigned fps_scaled = FPSNEWHI * 0x1000000ull / FPSNEWLO;
    unsigned long audio_samples = 0;
    unsigned remaining_delay = DELAY_BEGIN;

    long framewait = FIRST_FRAME;

    // Disable buffering for the streams to avoid stdio reading a piece
    // of the file on each successive seek. Also, we know exactly how
    // much we want to read each time, so stdio buffering is not helpful.
    if(ai && RATE) std::setbuf(ai, NULL);
    if(vi) std::setbuf(vi, NULL);

    for(long framecounter=0; framecounter<LAST_FRAME; ++framecounter)
    {
        const unsigned bytes_per_60_in_pixels = (IN_BPP == 15 ? 16 : IN_BPP) * 60 / 8;
        const unsigned bytes_per_60_out_pixels =(OUT_BPP== 15 ? 16 :OUT_BPP) * 60 / 8;
        const unsigned long in_image_size = WIDTH * HEIGHT * bytes_per_60_in_pixels / 60;
        const unsigned long out_image_size= WIDTH * HEIGHT * bytes_per_60_out_pixels / 60;
        char* VidBuf = (char*)
            std::malloc( in_image_size > out_image_size ? in_image_size : out_image_size );

        const unsigned bytes_per_60_samples = CHANNELS * 60 * BITS / 8;
        audio_samples += RATE * FPSLO;
        const unsigned take_audio_samples = audio_samples / FPSHI;
        const unsigned take_audio_bytes = bytes_per_60_samples/60*take_audio_samples;
        char* AudBuf = (char*)
            std::malloc( RATE * bytes_per_60_samples / 60 ); // Assume FPS is always >= 1.000

        audio_samples -= take_audio_samples * FPSHI;

        if(remaining_delay > 0)
        {
            std::memset(VidBuf, 0, in_image_size);
            std::memset(AudBuf, 0, take_audio_bytes);
            --remaining_delay;
        }
        else
        {
            if(IN_BPP == 12 || IN_BPP == 17)
            {
                std::fprintf(stderr, "hudmaker: Cannot scale chroma-supersampled images\n");
                std::abort();
            }
            const unsigned by = bytes_per_60_in_pixels / 60;
            const unsigned original_video_frame_size =
                ORIGWIDTH * ORIGHEIGHT * bytes_per_60_in_pixels / 60;

            if(framewait > 0)
            {
                if(ai && RATE) std::fseek(ai, take_audio_bytes, SEEK_CUR);
                if(vi) std::fseek(vi, original_video_frame_size, SEEK_CUR);
            }
            else
            {
                if(ai && RATE) std::fread(AudBuf, bytes_per_60_samples/60, take_audio_samples, ai);
                if(ORIGWIDTH == WIDTH && ORIGHEIGHT == HEIGHT)
                {
                    std::fread(VidBuf, 1, original_video_frame_size, vi);
                }
                else
                {
                    static std::vector<char> TempBuf( original_video_frame_size );
                    /* This buffer is "static" because otherwise, at
                     * high resolutions we can easily run out of stack space.
                     */
                    std::fread(&TempBuf[0], 1, original_video_frame_size, vi);
                    /* Point-scale the original image to the resulting image's size */

                    /*if((IN_BPP == 24
                     || IN_BPP == 32)
                    && WIDTH > ORIGWIDTH*2 && HEIGHT > ORIGHEIGHT*2)
                    {
                        std::vector<Triplet<unsigned char> > input_frame
                            ( ORIGWIDTH*2 * ORIGHEIGHT*2 );
                        std::vector<Triplet<LanczosFloat> >  scaled_frame;

                        //#pragma omp parallel for schedule(static) num_threads(3)
                        for(unsigned y=0; y<ORIGHEIGHT; ++y)
                        {
                            const unsigned char* src =
                                (const unsigned char*) &TempBuf[0]
                                + y*ORIGWIDTH*by;
                            unsigned p = (y*2) * (ORIGWIDTH*2);
                            for(unsigned x=0; x<ORIGWIDTH; ++x, p+=2, src+=by)
                            {
                                input_frame[p+1 +ORIGWIDTH*2] =
                                input_frame[p   +ORIGWIDTH*2] =
                                input_frame[p+1] =
                                input_frame[p] =
                                    Triplet<unsigned char>
                                        (src[0], src[1], src[2], (unsigned char)0);
                            }
                        }
                        LanczosScale(ORIGWIDTH*2, ORIGHEIGHT*2, WIDTH,HEIGHT,
                            input_frame, scaled_frame);
                        #pragma omp parallel for schedule(static) num_threads(3)
                        for(unsigned y=0; y<HEIGHT; ++y)
                        {
                            char* tgt = VidBuf + y*WIDTH*by;
                            unsigned p = y*WIDTH;
                            for(unsigned x=0; x<WIDTH; ++x,++p)
                            {
                                const int o16 = (x+4*y)%16;
                                int r = Quantize4x4<0xFF>     (o16, scaled_frame[p][0]); if(r<0)r=0; if(r>255)r=255;
                                int g = Quantize4x4<0xFF>     (o16, scaled_frame[p][1]); if(g<0)r=0; if(g>255)r=255;
                                int b = Quantize4x4<0xFF>     (o16, scaled_frame[p][2]); if(b<0)r=0; if(b>255)r=255;
                                tgt[0] = r;
                                tgt[1] = g;
                                tgt[2] = b;
                                tgt += by;
                            }
                        }
                    }
                    else
                    */
                    {
                        #pragma omp parallel for schedule(static) num_threads(3)
                        for(unsigned y=0; y< ORIGHEIGHT; ++y)
                        {
                            const unsigned tgt_y = y * HEIGHT / ORIGHEIGHT;
                            const unsigned yscale = (y+1)*HEIGHT / ORIGHEIGHT - tgt_y;

                            char* src = &TempBuf[0] + y*ORIGWIDTH*by;
                            char* tgt = VidBuf      + tgt_y*WIDTH*by;

                            unsigned xcarry = 0;
                            for(unsigned x=0; x< ORIGWIDTH; ++x)
                            {
                                xcarry += WIDTH;
                                const unsigned xscale = xcarry / ORIGWIDTH;
                                xcarry %= ORIGWIDTH;

                                unsigned xdiff = xscale*by;
                                for(unsigned ys=0, ysp=yscale*WIDTH; ys<ysp; ys+=WIDTH)
                                {
                                    char* x_tgt     = &tgt[by*ys];
                                    char* x_tgt_end = x_tgt + xdiff;
                                    while(x_tgt < x_tgt_end)
                                        { std::memcpy(x_tgt, src, by);
                                          x_tgt += by; }
                                }
                                /*for(unsigned c=0; c<by; ++c, ++tgt)
                                {
                                    char byt = *src++;
                                    for(unsigned ys=0, ysp=yscale*WIDTH; ys<ysp; ys += WIDTH)
                                    for(unsigned xs=0; xs<xscale; ++xs)
                                        tgt[by * (xs + ys)] = byt;
                                }*/
                                tgt += by * (xscale/*-1*/);
                                src += by;
                            }
                        }
                    } // lanczos condition
                } // scale condition
            } // frameskip condition
        } // delay condition

        if(framewait > 0)
        {
            --framewait;
            std::free(VidBuf);
            std::free(AudBuf);
        }
        else
        {
            static unsigned long long cache = 0;
            cache += FPSLO*FPSNEWHI;
            bool v_used = false;
            while(cache >= FPSNEWLO*FPSHI)
            {
                char* vframe = VidBuf;
                v_used = true;
                cache -= FPSNEWLO*FPSHI;
                if(cache >= FPSNEWLO*FPSHI) // Will this frame be duplicated?
                {
                    // If so, make a copy of VidBuf so that the
                    // next call to AddTo() won't deallocate it
                    // when it still will be needed
                    VidBuf = (char*)std::malloc(out_image_size);
                    memcpy(VidBuf, vframe, out_image_size);
                }
                ov.AddTo(vframe, framecounter);
                NESVideoLoggingVideoPtr(vframe, WIDTH, HEIGHT, fps_scaled, OUT_BPP);
            }
            NESVideoLoggingAudioPtr(AudBuf, RATE, BITS, CHANNELS, take_audio_samples);
            if(!v_used) std::free(VidBuf);

            if(std::feof(vi) || std::feof(ai)) break;
        }
    }
    std::fclose(ai);
    std::fclose(vi);
}
